"""The core part of mttf that can be imported without touching the Tensorflow package."""

import typing as tp
import yaml

__all__ = [
    "ModelSyntaxError",
    "ModelParams",
    "MHAParams",
    "MHAPool2DCascadeParams",
    "MobileNetV3MixerParams",
]


class ModelSyntaxError(SyntaxError):
    pass


class ModelParams(yaml.YAMLObject):
    """Parameters for defining and creating a model.

    This is an abstract class. The user should subclass from this class to define their own class
    which represents the collection of parameters to create models of a given family.

    Parameters
    ----------
    gen : int
        model generation/family number, starting from 1
    """

    yaml_tag = "!ModelParams"

    def __init__(self, gen: int = 1):
        self.gen = gen


class MHAParams(ModelParams):
    """Parameters for creating an MHA layer.

    Parameters
    ----------
    n_heads : int
        number of heads
    key_dim : int, optional
        dimensionality of each (projected) key/query vector. If not provided, it is set as the last
        dim of the query tensor integer-divided by `n_heads`.
    value_dim : int, optional
        dimensionality of each (projected) value vector. If not provided, it is set as `key_dim`.
    output_shape : object
        passed as-is to MultiHeadAttention
    gen : int
        model generation/family number, starting from 1
    """

    yaml_tag = "!MHAParams"

    def __init__(
        self,
        n_heads: int = 4,
        key_dim: tp.Optional[int] = None,
        value_dim: tp.Optional[int] = None,
        output_shape: object = None,
        gen: int = 1,
    ):
        super().__init__(gen=gen)

        self.n_heads = n_heads
        self.key_dim = key_dim
        self.value_dim = value_dim
        self.output_shape = output_shape

    def to_json(self):
        """Returns an equivalent json object."""
        return {
            "n_heads": self.n_heads,
            "key_dim": self.key_dim,
            "value_dim": self.value_dim,
            "output_shape": self.output_shape,
            "gen": self.gen,
        }

    @classmethod
    def from_json(cls, json_obj):
        """Instantiates from a json object."""
        return MHAParams(
            n_heads=json_obj["n_heads"],
            key_dim=json_obj.get("key_dim", None),
            value_dim=json_obj.get("value_dim", None),
            output_shape=json_obj.get("output_shape", None),
            gen=json_obj["gen"],
        )


class MHAPool2DCascadeParams(ModelParams):
    """Parameters for creating a cascade of MHAPool2D layers.

    The architecture is a cascade of K MHAPool2D layers followed by optionally a SimpleMHA2D layer.
    For a given number M which is the maximum number of MHAPool2D layers, the design is to cascade
    K <= M MHAPool2D layers such that either the grid resolution is 1x1 or K == M. If the grid
    resolution is not 1x1 at the K-th layer, then a SimpleMHA2D layer is cascaded to finish the
    job.

    All layer activations in the grid follow the same type, except for the last layer which can be
    modified.

    Parameters
    ----------
    n_heads : int
        number of heads
    expansion_factor : float
        expansion factor at each layer
    pooling : {'avg', 'max'}
        pooling type
    dropout : float
        dropout probability
    max_num_pooling_layers : int
        maximum number of pooling layers before a SimpleMHA2D layer to finish the job, if the grid
        resolution has not reached 1x1
    activation: str
        activation type for all pooling layers but maybe the last one
    final_activation : str
        activation type for the last layer, which can be MHAPool2D or SimpleMHA2D
    output_all : bool, optional
        If False, it returns the output tensor of the last layer. Otherwise, it additionally
        returns the output tensor of every attention layer before the last layer.
    gen : int
        model generation/family number, starting from 1
    """

    yaml_tag = "!MHAPool2DCascadeParams"

    def __init__(
        self,
        n_heads: int = 20,
        expansion_factor: float = 1.5,
        pooling: str = "max",
        dropout: float = 0.2,
        max_num_pooling_layers: int = 10,
        activation: str = "swish",
        final_activation: tp.Optional[str] = "swish",
        output_all: tp.Optional[bool] = False,
        gen: int = 1,
    ):
        super().__init__(gen=gen)

        self.n_heads = n_heads
        self.expansion_factor = expansion_factor
        self.pooling = pooling
        self.dropout = dropout
        self.max_num_pooling_layers = max_num_pooling_layers
        self.activation = activation
        self.final_activation = (
            activation if final_activation is None else final_activation
        )
        self.output_all = output_all

    def to_json(self):
        """Returns an equivalent json object."""
        return {
            "n_heads": self.n_heads,
            "expansion_factor": self.expansion_factor,
            "pooling": self.pooling,
            "dropout": self.dropout,
            "max_num_pooling_layers": self.max_num_pooling_layers,
            "activation": self.activation,
            "final_activation": self.final_activation,
            "output_all": self.output_all,
            "gen": self.gen,
        }

    @classmethod
    def from_json(cls, json_obj):
        """Instantiates from a json object."""
        return MHAPool2DCascadeParams(
            n_heads=json_obj["n_heads"],
            expansion_factor=json_obj["expansion_factor"],
            pooling=json_obj["pooling"],
            dropout=json_obj["dropout"],
            max_num_pooling_layers=json_obj["max_num_pooling_layers"],
            activation=json_obj["activation"],
            final_activation=json_obj["final_activation"],
            output_all=json_obj.get("output_all", False),
            gen=json_obj["gen"],
        )


class MobileNetV3MixerParams(ModelParams):
    """Parameters for creating a MobileNetV3Mixer.

    Parameters
    ----------
    variant : {'mobilenet', 'maxpool', 'mhapool'}
        Variant of the mixer block. The output tensor has 1x1 spatial resolution. If 'mobilenet' is
        specified, the mixer follows 'mobilenet' style, including mainly 2 Conv layers and one
        GlobalAveragePooling2D layer. If 'maxpool' is specified, grid processing is just a
        GlobalMaxPool2D layer. If 'mhapool' is used, a cascade of MHAPool2D layers is used until
        the last layer outputs a 1x1 tensor.
    mhapool_cascade_params : mt.tfc.MHAPool2DCascadeParams, optional
        The parameters defining a cascade of MHAPool2D layers. Only valid for 'mhapool' mixer type.
    gen : int
        model generation/family number, starting from 1
    """

    yaml_tag = "!MobileNetV3MixerParams"

    def __init__(
        self,
        variant: str = "mobilenet",
        mhapool_cascade_params: tp.Optional[MHAPool2DCascadeParams] = None,
        gen: int = 1,
    ):
        super().__init__(gen=gen)

        self.variant = variant
        self.mhapool_cascade_params = mhapool_cascade_params

    def to_json(self):
        """Returns an equivalent json object."""
        if self.mhapool_cascade_params is None:
            mhapool_params = None
        else:
            mhapool_params = self.mhapool_cascade_params.to_json()
        return {
            "variant": self.variant,
            "mhapool_cascade_params": mhapool_params,
            "gen": self.gen,
        }

    @classmethod
    def from_json(cls, json_obj):
        """Instantiates from a json object."""
        mhapool_params = json_obj.get("mhapool_cascade_params", None)
        if mhapool_params is not None:
            mhapool_params = MHAPool2DCascadeParams.from_json(mhapool_params)
        return MobileNetV3MixerParams(
            variant=json_obj["variant"],
            mhapool_cascade_params=mhapool_params,
            gen=json_obj["gen"],
        )
